from dataclasses import dataclass
from enum import Enum

from dataclasses_json import dataclass_json

from ..._helpers import extract_uint32, extract_int32, extract_uint16


class ReaderChipVersion(Enum):
    AS3992 = 1
    AS3993 = 2


class ModuleType(Enum):
    NUR_05W = 1
    XNUR_05W = 2
    NUR_L2_05W = 3
    NUR_L2_1W = 4


@dataclass_json
@dataclass
class NurDeviceCaps:
    dwSize: int = None
    flagSet1: int = None
    flagSet2: int = None
    maxTxdBm: int = None
    txAttnStep: int = None
    maxTxmW: int = None
    txSteps: int = None
    szTagBuffer: int = None
    curCfgMaxAnt: int = None
    curCfgMaxGPIO: int = None
    chipVersion: ReaderChipVersion = None
    moduleType: ModuleType = None
    moduleConfigFlags: int = None


def parse_get_device_capabilities_response(payload: bytearray):
    device_caps = NurDeviceCaps()
    device_caps.dwSize = extract_uint32(payload)
    device_caps.flagSet1 = extract_uint32(payload)
    device_caps.flagSet2 = extract_uint32(payload)
    device_caps.maxTxdBm = extract_int32(payload)
    device_caps.txAttnStep = extract_int32(payload)
    device_caps.maxTxmW = extract_uint16(payload)
    device_caps.txSteps = extract_uint16(payload)
    device_caps.szTagBuffer = extract_uint16(payload)
    device_caps.curCfgMaxAnt = extract_uint16(payload)
    device_caps.curCfgMaxGPIO = extract_uint16(payload)
    device_caps.chipVersion = ReaderChipVersion(extract_uint16(payload))
    device_caps.moduleType = ModuleType(extract_uint16(payload))
    device_caps.moduleConfigFlags = extract_uint32(payload)
    return device_caps



import logging
import queue
import struct
import time
from threading import Thread
from typing import Callable, List

from . import Packet
from .command.get_device_capabilities import parse_get_device_capabilities_response
from .command.get_id_buffer import parse_get_id_buffer_response
from .command.get_id_buffer_meta import parse_get_id_buffer_meta_response, NurTagDataMeta
from .command.get_mode import parse_get_mode_response
from .command import HeaderFlagCodes, StatusCode
from .command.inventory import parse_inventory_response
from .command.inventory_stream import parse_inventory_stream_notification, InventoryStreamNotification
from .command.module_setup import parse_module_setup_response
from .command.get_reader_info import parse_get_reader_info_response
from .. import CommandCode

logger = logging.getLogger(__name__)


class RxHandler:

    def __init__(self):
        self.buffer = bytearray()
        self.response_queue = queue.Queue()
        self.notification_queue = queue.Queue()
        self.notification_callback = None
        self._callback_thread = Thread(target=self._callback_thread_fxn, daemon=True, name='CallbackThread')
        self._callback_thread.start()

    def set_notification_callback(self, callback: Callable[[InventoryStreamNotification,
                                                            List[NurTagDataMeta]], None]) -> None:
        self.notification_callback = callback

    def _callback_thread_fxn(self) -> None:
        while True:
            if not self.notification_queue.empty():
                [notification, tags] = self.notification_queue.get()
                if self.notification_callback:
                    self.notification_callback(notification, tags)
            time.sleep(0.001)

    def append_data(self, data):
        if data is not None:
            self.buffer += data
            self._try_parse_data()

    def _try_parse_data(self):
        try:
            # Discard data until PREAMBLE
            start = self.buffer.find(bytearray([Packet.START_BYTE]))
            if start > 0:
                logger.warning('Syching PREAMBLE.')
                self.buffer = self.buffer[start:]

            # Check if Header available
            while len(self.buffer) > 5:
                # Checksum validation
                checksum_rx = self.buffer[5]
                checksum_calc = Packet.checksum(self.buffer[0:5])
                if checksum_calc != checksum_rx:
                    logger.warning('Checksum mismatch')
                    # Remove START_BYTE
                    del self.buffer[0]
                else:
                    payload_len = struct.unpack('<H', self.buffer[1:3])[0]
                    header_flags = struct.unpack('<H', self.buffer[3:5])[0]
                    # Check if Full packet available
                    if len(self.buffer) > 5 + payload_len:
                        # Check CRC
                        crc_rx = struct.unpack('<H', self.buffer[7 + payload_len - 3: 7 + payload_len])[0]
                        crc_calc = Packet.crc16(self.buffer[6:7 + payload_len - 3])
                        if crc_calc != crc_rx:
                            logger.warning('CRC mismatch')
                        else:
                            # Extract data
                            command = CommandCode(self.buffer[6])
                            payload = self.buffer[7:7 + payload_len - 3]

                            status = StatusCode(payload[0])
                            if status != StatusCode.SUCCESS:
                                logger.error(status)
                                return
                            payload = payload[1:]

                            # Process message
                            if header_flags == HeaderFlagCodes.RESPONSE.value:
                                self._process_response(command, payload)
                            elif header_flags & HeaderFlagCodes.NOTIFICATION.value:
                                self._process_notification(command, payload)

                        # Remove processed data
                        del self.buffer[0:8 + payload_len]
                    else:
                        # Missing payload data
                        break
        except Exception as e:
            logger.error(e)

    def _process_response(self, command: CommandCode, payload: bytearray):
        if command is CommandCode.PING:
            self.response_queue.put(True)
            return
        if command is CommandCode.RESET:
            self.response_queue.put(True)
            return
        if command is CommandCode.RESTART:
            self.response_queue.put(True)
            return
        if command is CommandCode.GET_MODE:
            self.response_queue.put(parse_get_mode_response(payload))
            return
        if command is CommandCode.GET_READER_INFO:
            self.response_queue.put(parse_get_reader_info_response(payload))
            return
        if command is CommandCode.GET_DEVICE_CAPABILITIES:
            self.response_queue.put(parse_get_device_capabilities_response(payload))
            return
        if command is CommandCode.GET_MODULE_SETUP:
            self.response_queue.put(parse_module_setup_response(payload))
            return
        if command is CommandCode.SIMPLE_INVENTORY:
            self.response_queue.put(parse_inventory_response(payload))
            return
        if command is CommandCode.GET_ID_BUFFER:
            self.response_queue.put(parse_get_id_buffer_response(payload))
            return
        if command is CommandCode.GET_ID_BUFFER_META:
            self.response_queue.put(parse_get_id_buffer_meta_response(payload))
            return
        if command is CommandCode.CLEAR_ID_BUFFER:
            self.response_queue.put(True)
            return
        if command is CommandCode.INVENTORY_STREAM:
            self.response_queue.put(True)
            return

    def _process_notification(self, command: CommandCode, payload: bytearray):
        if command is CommandCode.NOTIFICATION_INVENTORY:
            self.notification_queue.put(parse_inventory_stream_notification(payload))
            return

    def get_response(self, timeout=1):
        start = time.monotonic()
        while self.response_queue.empty():
            time.sleep(0.001)
            if time.monotonic() - start > timeout:
                raise TimeoutError
        return self.response_queue.get()



import struct

from ..._helpers import (extract_uint32, extract_uint8, extract_uint16, extract_int8,
                                 to_uint8_bytes, to_uint32_bytes, to_uint16_bytes, to_int8_bytes)
from dataclasses import dataclass
from enum import Enum
from typing import List

from dataclasses_json import dataclass_json


class ModuleSetupFlags(Enum):
    LINKFREQ = (1 << 0)  # linkFreq field in struct NUR_MODULESETUP is valid */
    RXDEC = (1 << 1)  # rxDecoding field in struct NUR_MODULESETUP is valid */
    TXLEVEL = (1 << 2)  # txLevel field in struct NUR_MODULESETUP is valid */
    TXMOD = (1 << 3)  # txModulation field in struct NUR_MODULESETUP is valid */
    REGION = (1 << 4)  # regionId field in struct NUR_MODULESETUP is valid */
    INVQ = (1 << 5)  # inventoryQ field in struct NUR_MODULESETUP is valid */
    INVSESSION = (1 << 6)  # inventorySession field in struct NUR_MODULESETUP is valid */
    INVROUNDS = (1 << 7)  # inventoryRounds field in struct NUR_MODULESETUP is valid */
    ANTMASK = (1 << 8)  # antennaMask field in struct NUR_MODULESETUP is valid */
    SCANSINGLETO = (1 << 9)  # scanSingleTriggerTimeout field in struct NUR_MODULESETUP is valid */
    INVENTORYTO = (1 << 10)  # inventoryTriggerTimeout field in struct NUR_MODULESETUP is valid */
    SELECTEDANT = (1 << 11)  # selectedAntenna field in struct NUR_MODULESETUP is valid */
    OPFLAGS = (1 << 12)  # opFlags field in struct NUR_MODULESETUP is valid */
    INVTARGET = (1 << 13)  # inventoryTarget field in struct NUR_MODULESETUP is valid */
    INVEPCLEN = (1 << 14)  # inventoryEpcLength field in struct NUR_MODULESETUP is valid */
    READRSSIFILTER = (1 << 15)  # readRssiFilter field in struct NUR_MODULESETUP is valid */
    WRITERSSIFILTER = (1 << 16)  # writeRssiFilter field in struct NUR_MODULESETUP is valid */
    INVRSSIFILTER = (1 << 17)  # inventoryRssiFilter field in struct NUR_MODULESETUP is valid */
    READTIMEOUT = (1 << 18)  # readTO field in struct NUR_MODULESETUP is valid */
    WRITETIMEOUT = (1 << 19)  # writeTO field in struct NUR_MODULESETUP is valid */
    LOCKTIMEOUT = (1 << 20)  # lockTO field in struct NUR_MODULESETUP is valid */
    KILLTIMEOUT = (1 << 21)  # killTO field in struct NUR_MODULESETUP is valid */
    AUTOPERIOD = (1 << 22)  # stixPeriod field in struct NUR_MODULESETUP is valid */
    PERANTPOWER = (1 << 23)  # antPower field in struct NUR_MODULESETUP is valid */
    PERANTOFFSET = (1 << 24)  # powerOffset field in struct NUR_MODULESETUP is valid */
    ANTMASKEX = (1 << 25)  # antennaMaskEx field in struct NUR_MODULESETUP is valid */
    AUTOTUNE = (1 << 26)  # autotune field in struct NUR_MODULESETUP is valid */
    PERANTPOWER_EX = (1 << 27)  # antPowerEx field in struct NUR_MODULESETUP is valid */
    RXSENS = (1 << 28)  # rxSensitivity field in struct NUR_MODULESETUP is valid */

    ALL = ((1 << 29) - 1)  # All setup flags in the structure. */

    # ADDED NUR2 7.0
    #RFPROFILE = (1 << 29)  # rfProfile field in struct NUR_MODULESETUP is valid */

    # ADDED NUR2 7.5, NanoNur 10.2
    #TO_SLEEP_TIME = (1 << 30)  # toSleepTime field in struct NUR_MODULESETUP is valid */

    #ALL_NUR2 = ((1 << 31) - 1)  # All setup flags in the structure. */


class ModuleSetupLinkFreq(Enum):
    BLF_160 = 160000
    BLF_256 = 256000
    BLF_320 = 320000


class ModuleSetupRxDec(Enum):
    FM0 = 0
    MILLER_2 = 1
    MILLER_4 = 2
    MILLER_8 = 3


class ModuleSetupTxMod(Enum):
    ASK = 0
    PRASK = 1


class ModuleSetupRegion(Enum):
    EU = 0
    FCC = 1
    PRC = 2
    Malaysia = 3
    Brazil = 4
    Australia = 5
    NewZealand = 6
    Japan_250mW_LBT = 7
    Japan_500mW_DRM = 8
    Korea_LBT = 9
    India = 10
    Russia = 11
    Vietnam = 12
    Singapore = 13
    Thailand = 14
    Philippines = 15
    Morocco = 16
    Peru = 17


class ModuleSetupInvTarget(Enum):
    A = 0
    B = 1
    AB = 2


class ModuleSetupPowerSave(Enum):
    NOT_IN_USE = 0
    MAX_1000MS_QUIET_BETWEEN_INVENTORIES = 1
    MAX_500MS_QUIET_BETWEEN_INVENTORIES = 2
    MAX_100MS_QUIET_BETWEEN_INVENTORIES = 3


class ModuleSetupRxSens(Enum):
    NOMINAL = 0
    LOW = 1
    HIGH = 2


class NurRssiFilter:
    min: int = None
    max: int = None


class NurAutoTuneSetup:
    mode: int = None
    threshold_dBm: int = None


@dataclass_json
@dataclass
class ModuleSetup:
    link_freq: ModuleSetupLinkFreq = None
    rx_decoding: ModuleSetupRxDec = None
    tx_level: int = None
    tx_modulation: ModuleSetupTxMod = None
    region_id: ModuleSetupRegion = None
    inventory_q: int = None
    inventory_session: int = None
    inventory_rounds: int = None
    antenna_mask: int = None
    scan_single_trigger_timeout: int = None
    inventory_trigger_timeout: int = None
    selected_antenna: int = None
    op_flags: int = None
    inventory_target: ModuleSetupInvTarget = None
    inventory_epc_length: int = None
    read_rssi_filter = None
    write_rssi_filter: NurRssiFilter = None
    inventory_rssi_filter: NurRssiFilter = None
    read_to: int = None
    write_to: int = None
    lock_to: int = None
    kill_to: int = None
    period_setup: ModuleSetupPowerSave = None
    ant_power: List[int] = None
    power_offset: List[int] = None
    antenna_mask_ex: int = None
    autotune: NurAutoTuneSetup = None
    ant_power_ex: List[int] = None
    rx_sensitivity: ModuleSetupRxSens = None
    #rf_profile: int = None
    #to_sleep_time: int = None


def parse_module_setup_response(payload: bytearray):
    module_setup = ModuleSetup()
    setup_flags = extract_uint32(payload)
    if setup_flags & ModuleSetupFlags.LINKFREQ.value:
        module_setup.link_freq = ModuleSetupLinkFreq(extract_uint32(payload))
    if setup_flags & ModuleSetupFlags.RXDEC.value:
        module_setup.rx_decoding = ModuleSetupRxDec(extract_uint8(payload))
    if setup_flags & ModuleSetupFlags.TXLEVEL.value:
        module_setup.tx_level = extract_uint8(payload)
    if setup_flags & ModuleSetupFlags.TXMOD.value:
        module_setup.tx_modulation = ModuleSetupTxMod(extract_uint8(payload))
    if setup_flags & ModuleSetupFlags.REGION.value:
        module_setup.region_id = ModuleSetupRegion(extract_uint8(payload))
    if setup_flags & ModuleSetupFlags.INVQ.value:
        module_setup.inventory_q = extract_uint8(payload)
    if setup_flags & ModuleSetupFlags.INVSESSION.value:
        module_setup.inventory_session = extract_uint8(payload)
    if setup_flags & ModuleSetupFlags.INVROUNDS.value:
        module_setup.inventory_rounds = extract_uint8(payload)
    if setup_flags & ModuleSetupFlags.ANTMASK.value:
        module_setup.antenna_mask = extract_uint8(payload)
    if setup_flags & ModuleSetupFlags.SCANSINGLETO.value:
        module_setup.scan_single_trigger_timeout = extract_uint16(payload)
    if setup_flags & ModuleSetupFlags.INVENTORYTO.value:
        module_setup.inventory_trigger_timeout = extract_uint16(payload)
    if setup_flags & ModuleSetupFlags.SELECTEDANT.value:
        module_setup.selected_antenna = extract_uint8(payload)
    if setup_flags & ModuleSetupFlags.OPFLAGS.value:
        module_setup.op_flags = extract_uint32(payload)
    if setup_flags & ModuleSetupFlags.INVTARGET.value:
        module_setup.inventory_target = ModuleSetupInvTarget(extract_uint8(payload))
    if setup_flags & ModuleSetupFlags.INVEPCLEN.value:
        module_setup.inventory_epc_length = extract_uint8(payload)
    if setup_flags & ModuleSetupFlags.READRSSIFILTER.value:
        module_setup.read_rssi_filter = NurRssiFilter()
        module_setup.read_rssi_filter.min = extract_int8(payload)
        module_setup.read_rssi_filter.max = extract_int8(payload)
    if setup_flags & ModuleSetupFlags.WRITERSSIFILTER.value:
        module_setup.write_rssi_filter = NurRssiFilter()
        module_setup.write_rssi_filter.min = extract_int8(payload)
        module_setup.write_rssi_filter.max = extract_int8(payload)
    if setup_flags & ModuleSetupFlags.INVRSSIFILTER.value:
        module_setup.inventory_rssi_filter = NurRssiFilter()
        module_setup.inventory_rssi_filter.min = extract_int8(payload)
        module_setup.inventory_rssi_filter.max = extract_int8(payload)
    if setup_flags & ModuleSetupFlags.READTIMEOUT.value:
        module_setup.read_to = extract_uint16(payload)
    if setup_flags & ModuleSetupFlags.WRITETIMEOUT.value:
        module_setup.write_to = extract_uint16(payload)
    if setup_flags & ModuleSetupFlags.LOCKTIMEOUT.value:
        module_setup.lock_to = extract_uint16(payload)
    if setup_flags & ModuleSetupFlags.KILLTIMEOUT.value:
        module_setup.kill_to = extract_uint16(payload)
    if setup_flags & ModuleSetupFlags.AUTOPERIOD.value:
        module_setup.period_setup = ModuleSetupPowerSave(extract_uint8(payload))
    if setup_flags & ModuleSetupFlags.PERANTPOWER.value:
        module_setup.ant_power = []
        for i in range(4):
            module_setup.ant_power.append(extract_int8(payload))
    if setup_flags & ModuleSetupFlags.PERANTOFFSET.value:
        module_setup.power_offset = []
        for i in range(4):
            module_setup.power_offset.append(extract_int8(payload))
    if setup_flags & ModuleSetupFlags.ANTMASKEX.value:
        module_setup.antenna_mask_ex = extract_uint32(payload)
    if setup_flags & ModuleSetupFlags.AUTOTUNE.value:
        module_setup.autotune = NurAutoTuneSetup()
        module_setup.autotune.mode = extract_uint8(payload)
        module_setup.autotune.threshold_dBm = extract_uint8(payload)
    if setup_flags & ModuleSetupFlags.PERANTPOWER_EX.value:
        module_setup.ant_power_ex = []
        for i in range(32):
            module_setup.ant_power_ex.append(extract_int8(payload))
    if setup_flags & ModuleSetupFlags.RXSENS.value:
        dummy_byte = extract_int8(payload)
        module_setup.rx_sensitivity = ModuleSetupRxSens(extract_int8(payload))

    return module_setup


def populate_module_setup_args(setup_flags: int, module_setup: ModuleSetup):
    args = [struct.pack('I', setup_flags)]
    if setup_flags & ModuleSetupFlags.LINKFREQ.value:
        args.append(to_uint32_bytes(module_setup.link_freq.value))
    if setup_flags & ModuleSetupFlags.RXDEC.value:
        args.append(to_uint8_bytes(module_setup.rx_decoding.value))
    if setup_flags & ModuleSetupFlags.TXLEVEL.value:
        args.append(to_uint8_bytes(module_setup.tx_level))
    if setup_flags & ModuleSetupFlags.TXMOD.value:
        args.append(to_uint8_bytes(module_setup.tx_modulation.value))
    if setup_flags & ModuleSetupFlags.REGION.value:
        args.append(to_uint8_bytes(module_setup.region_id.value))
    if setup_flags & ModuleSetupFlags.INVQ.value:
        args.append(to_uint8_bytes(module_setup.inventory_q))
    if setup_flags & ModuleSetupFlags.INVSESSION.value:
        args.append(to_uint8_bytes(module_setup.inventory_session))
    if setup_flags & ModuleSetupFlags.INVROUNDS.value:
        args.append(to_uint8_bytes(module_setup.inventory_rounds))
    if setup_flags & ModuleSetupFlags.ANTMASK.value:
        args.append(to_uint8_bytes(module_setup.antenna_mask))
    if setup_flags & ModuleSetupFlags.SCANSINGLETO.value:
        args.append(to_uint16_bytes(module_setup.scan_single_trigger_timeout))
    if setup_flags & ModuleSetupFlags.INVENTORYTO.value:
        args.append(to_uint16_bytes(module_setup.inventory_trigger_timeout))
    if setup_flags & ModuleSetupFlags.SELECTEDANT.value:
        args.append(to_uint8_bytes(module_setup.selected_antenna))
    if setup_flags & ModuleSetupFlags.OPFLAGS.value:
        args.append(to_uint32_bytes(module_setup.op_flags))
    if setup_flags & ModuleSetupFlags.INVTARGET.value:
        args.append(to_uint8_bytes(module_setup.inventory_target.value))
    if setup_flags & ModuleSetupFlags.INVEPCLEN.value:
        args.append(to_uint8_bytes(module_setup.inventory_epc_length))
    if setup_flags & ModuleSetupFlags.READRSSIFILTER.value:
        args.append(to_uint8_bytes(module_setup.read_rssi_filter.min))
        args.append(to_uint8_bytes(module_setup.read_rssi_filter.max))
    if setup_flags & ModuleSetupFlags.WRITERSSIFILTER.value:
        args.append(to_uint8_bytes(module_setup.write_rssi_filter.min))
        args.append(to_uint8_bytes(module_setup.write_rssi_filter.max))
    if setup_flags & ModuleSetupFlags.INVRSSIFILTER.value:
        args.append(to_uint8_bytes(module_setup.inventory_rssi_filter.min))
        args.append(to_uint8_bytes(module_setup.inventory_rssi_filter.max))
    if setup_flags & ModuleSetupFlags.READTIMEOUT.value:
        args.append(to_uint16_bytes(module_setup.read_to))
    if setup_flags & ModuleSetupFlags.WRITETIMEOUT.value:
        args.append(to_uint16_bytes(module_setup.write_to))
    if setup_flags & ModuleSetupFlags.LOCKTIMEOUT.value:
        args.append(to_uint16_bytes(module_setup.lock_to))
    if setup_flags & ModuleSetupFlags.KILLTIMEOUT.value:
        args.append(to_uint16_bytes(module_setup.kill_to))
    if setup_flags & ModuleSetupFlags.AUTOPERIOD.value:
        args.append(to_uint16_bytes(module_setup.period_setup.value))
    if setup_flags & ModuleSetupFlags.PERANTPOWER.value:
        for i in range(4):
            args.append(to_int8_bytes(module_setup.ant_power[i]))
    if setup_flags & ModuleSetupFlags.PERANTOFFSET.value:
        for i in range(4):
            args.append(to_int8_bytes(module_setup.power_offset[i]))
    if setup_flags & ModuleSetupFlags.ANTMASKEX.value:
        args.append(to_uint32_bytes(module_setup.antenna_mask_ex))
    if setup_flags & ModuleSetupFlags.AUTOTUNE.value:
        args.append(to_uint8_bytes(module_setup.autotune.mode))
        args.append(to_uint8_bytes(module_setup.autotune.threshold_dBm))
    if setup_flags & ModuleSetupFlags.PERANTPOWER_EX.value:
        for i in range(32):
            args.append(to_int8_bytes(module_setup.ant_power_ex[i]))
    if setup_flags & ModuleSetupFlags.RXSENS.value:
        args.append(0xFF)
        args.append(to_int8_bytes(module_setup.rx_sensitivity.value))

    return args



import struct
from enum import Enum


class HeaderFlagCodes(Enum):
    RESPONSE = 0x0000
    NOTIFICATION = 0x0001
    INVENTORY_READ_NOTIFICATION = 0x0002


class CommandCode(Enum):
    PING = 1
    RESET = 3
    GET_MODE = 4
    CLEAR_ID_BUFFER = 5
    GET_ID_BUFFER = 6
    GET_ID_BUFFER_META = 7
    GET_READER_INFO = 9
    GET_DEVICE_CAPABILITIES = 0xB
    RESTART = 0x14
    GET_MODULE_SETUP = 0x22
    SIMPLE_INVENTORY = 0x31
    INVENTORY_STREAM = 0x39
    NOTIFICATION_INVENTORY = 0x82


class StatusCode(Enum):
    SUCCESS = 0
    INVALID_COMMAND = 1
    INVALID_LENGTH = 2
    PARAMETER_OUT_OF_RANGE = 3
    INVALID_PARAMETER = 5
    CRC_ERROR = 11
    APPLICATION_NOT_PRESENT = 14



from enum import Enum
from typing import List, Any

from crccheck.crc import Crc16Ibm3740

from .command import CommandCode


class Packet:
    START_BYTE = 0xA5

    def __init__(self, command_code: CommandCode, args: List[Any]):
        self.command_buffer = bytearray()
        self.command_buffer.append(Packet.START_BYTE)
        self.command_buffer += int(0).to_bytes(2, 'little')  # Fake length
        self.command_buffer += int(0).to_bytes(2, 'little')  # Flag word
        self.command_buffer.append(0x00)  # Fake checksum
        self.command_buffer.append(command_code.value)
        arg_length = 0
        for arg in args:
            if isinstance(arg, list):
                for item in arg:
                    self.command_buffer.append(item)
                    arg_length += 1
            elif isinstance(arg, bytes):
                for item in arg:
                    self.command_buffer.append(item)
                    arg_length += 1
            elif isinstance(arg, bytearray):
                self.command_buffer += bytearray(arg)
                arg_length += len(arg)
            elif isinstance(arg, Enum):
                self.command_buffer.append(arg.value)
                arg_length += 1
            else:
                self.command_buffer.append(arg)
                arg_length += 1

        # Get CRC16 without header
        crc16 = self.crc16(self.command_buffer[6:])
        self.command_buffer += int(crc16).to_bytes(2, 'little')

        # Set real command payload length (without 6byte header)
        self.command_buffer[1:3] = (len(self.command_buffer) - 6).to_bytes(2, 'little')
        # Set real checksum
        self.command_buffer[5] = Packet.checksum(self.command_buffer[:5])

    def bytes(self):
        return self.command_buffer

    def get_command_code(self) -> CommandCode | None:
        if len(self.command_buffer) < 6:
            return None
        return CommandCode(self.command_buffer[6])

    def __str__(self):
        return self.command_buffer.hex(sep=' ').upper()

    @staticmethod
    def checksum(data: bytes):
        cs = 0xFF
        for byte in data:
            cs ^= byte
        return cs & 0xFF

    @staticmethod
    def crc16(data: bytes):
        crc = Crc16Ibm3740.calc(data)
        return crc



import struct


def extract_bytes(payload: bytearray, number_of_bytes: int) -> bytearray:
    b = payload[:number_of_bytes]
    del payload[:number_of_bytes]
    return b


def extract_string(payload: bytearray) -> str:
    string_length = payload[0]
    content = payload[1:string_length + 1]
    del payload[:string_length + 1]
    return content.decode('utf-8')


def extract_int8(payload: bytearray) -> int:
    i = struct.unpack('b', bytes([payload[0]]))[0]
    del payload[0]
    return i


def to_int8_bytes(value: int) -> bytes:
    return struct.pack('b', value)


def extract_uint8(payload: bytearray) -> int:
    i = struct.unpack('B', bytes([payload[0]]))[0]
    del payload[0]
    return i


def to_uint8_bytes(value: int) -> bytes:
    return struct.pack('B', value)


def extract_int16(payload: bytearray) -> int:
    i = struct.unpack('<h', payload[:2])[0]
    del payload[:2]
    return i


def to_int16_bytes(value: int) -> bytes:
    return struct.pack('<h', value)


def extract_uint16(payload: bytearray) -> int:
    i = struct.unpack('<H', payload[:2])[0]
    del payload[:2]
    return i


def to_uint16_bytes(value: int) -> bytes:
    return struct.pack('<H', value)


def extract_int32(payload: bytearray) -> int:
    i = struct.unpack('<i', payload[:4])[0]
    del payload[:4]
    return i


def to_int32_bytes(value: int) -> bytes:
    return struct.pack('<i', value)


def extract_uint32(payload: bytearray) -> int:
    i = struct.unpack('<I', payload[:4])[0]
    del payload[:4]
    return i


def to_uint32_bytes(value: int) -> bytes:
    return struct.pack('<I', value)